/*-
 * Copyright (c) 2000, 2002 Roman V. Palagin
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] = 
  "$Id: ncache.c,v 1.5 2002/10/08 11:22:57 romanp Exp $";
#endif

#include <sys/types.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/namei.h>
#include <sys/vnode.h>
#define _KERNEL
#include <sys/mount.h>
#undef _KERNEL
#include <ufs/ufs/quota.h>
#include <ufs/ufs/inode.h>

#include <nlist.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <osreldate.h>

#include "pmap.h"

/* from /sys/kern/vfs_cache.c */
struct	namecache {
	LIST_ENTRY(namecache) nc_hash;	/* hash chain */
	LIST_ENTRY(namecache) nc_src;	/* source vnode list */
	TAILQ_ENTRY(namecache) nc_dst;	/* destination vnode list */
	struct	vnode *nc_dvp;		/* vnode of parent of name */
	struct	vnode *nc_vp;		/* vnode the name refers to */
	u_char	nc_flag;		/* flag bits */
	u_char	nc_nlen;		/* length of name */
	char	nc_name[0];		/* segment name */
};

static struct nlist nl[] = {
#define	N_NCHASH	0
    { "_nchash", 0},
#define N_NCHASHTBL	1
    { "_nchashtbl", 0 },
    { 0,0 },
};

/* hash is based on (vn, dvn) */
#define LNCHASH(i,n)		(((((int)(n)>>2)+((int)(i)))*31415) % nchash)

struct l_nchash 	*L_nchash, *L_nc;
u_int 			nchash, h_idx;

static struct l_nch 	*namec_addr(u_long, u_long);
static void 		make_pa_links(void);

int 
namec_load(void)
{
	int i, error; 
	u_long ptr;
	LIST_HEAD(nchashhead, namecache) *nchashtbl;
	struct namecache *nc;
	struct vnode vn, dvn;
	struct l_nch *l_nch;

	k_nlist(nl);
	k_read(nl[N_NCHASH].n_value, &nchash, sizeof(nchash));
	nchash++;

	/* Allocate memory for local hash */
	L_nchash = (struct l_nchash *)calloc(nchash, sizeof *L_nchash);
	
	nchashtbl = malloc(nchash * sizeof *nchashtbl);
	nc = malloc(sizeof *nc + MAXNAMLEN+1);
	k_read(nl[N_NCHASHTBL].n_value, &ptr, sizeof(ptr));
	k_read(ptr, nchashtbl, nchash * sizeof(*nchashtbl));
	for (i=0; i < nchash; i++) {
		ptr = (u_long)LIST_FIRST(nchashtbl+i);
		while (ptr) {
			error = k_read(ptr, nc, sizeof(*nc) + MAXNAMLEN);
			if (error == -1) {
				fprintf(stderr, "error reading from slot %d (from %d), ptr = 0x%08x)\n", i, nchash, ptr);
				goto empty_vnode;
			}
			if (nc->nc_vp == NULL)
				goto empty_vnode;
			k_read((u_long)nc->nc_dvp, &dvn, sizeof(dvn));
			k_read((u_long)nc->nc_vp, &vn, sizeof(vn)); 
			nc->nc_name[nc->nc_nlen] = '\0';
			
			/* Add node to local hash */
			h_idx = LNCHASH(nc->nc_vp, vn.v_id);
			L_nc = L_nchash+h_idx;
			if (!L_nc->n_idx) /* first entry in this bucket */
				L_nc->l_nch = malloc(sizeof(struct l_nch));
			else 
				L_nc->l_nch = realloc(L_nc->l_nch, sizeof(struct l_nch)*(L_nc->n_idx+1));
			l_nch = L_nc->l_nch+L_nc->n_idx;
			l_nch->vp = (u_long) nc->nc_vp; 
			l_nch->dvp = (u_long) nc->nc_dvp;
			l_nch->id = vn.v_id; 
			l_nch->did = dvn.v_id;
			l_nch->vp_type = vn.v_type; 
			l_nch->dvp_type = dvn.v_type;
			l_nch->flag = vn.v_flag; 
			l_nch->dflag = dvn.v_flag;
			strcpy(l_nch->name, nc->nc_name);
			l_nch->pa = NULL;
			L_nc->n_idx++;
		empty_vnode:			
			ptr = (u_long)LIST_NEXT(nc, nc_hash);
		}
	}
#ifdef DBG_NCHASH
	check_nchash();
#endif
	make_pa_links();
	return 0;
}

char *
namec_lookup(u_long va, u_long id)
{
	struct l_nch *p = namec_addr(va, id);
	char tmpnam[1025];
	static char Name[1024];
	struct mount mp;
	struct vnode vn;

	if (p) {
		/* l_nch emtry found, try to find up parent(s) name */
		strcpy(Name, p->name);
		for (p = p->pa; p ; p = p->pa) {
			strcpy(tmpnam, Name);
			sprintf(Name,"%s/", p->name);
			strcat(Name, tmpnam);
			if (p->dflag & VROOT) { /* top of FS reached */
				if (p->dvp) {
					k_read((u_long) p->dvp, &vn, sizeof(vn));
					if (vn.v_mount) {
						int rc = k_read((u_long) vn.v_mount, &mp, sizeof(mp));
			
						if (rc != -1) {
							strcpy(tmpnam, Name);
							sprintf(Name, "%s/", mp.mnt_stat.f_mntonname);
							strcat(Name, tmpnam);
						}
					}
				}
				break;
			}
		}
	}
	else strcpy(Name,"#lookup error#");

	return Name;
}

static void 
make_pa_links(void)
{
	register int i, k;
	
	for (i=0;i<nchash;i++) {
		if (!L_nchash[i].n_idx) continue;
		for (k=0;k<L_nchash[i].n_idx;k++) {
			struct l_nch *pa, *p;

			p = &L_nchash[i].l_nch[k];
			pa = namec_addr(p->dvp, p->did);
			p->pa = pa;
		}
	}
}

static struct l_nch *
namec_addr(u_long vp, u_long id)
{
	u_int idx = LNCHASH(vp, id);
	struct l_nch *hp = L_nchash[idx].l_nch;
	register int i;
	
	for (i=0 ; i<L_nchash[idx].n_idx ; i++,hp++)
		if ((hp->vp == vp) && (hp->id == id))
			return  hp;
	return (struct l_nch *) 0;
}
